#!/usr/bin/env python

import sys
import os
import time
from numpy import *


__version__ = "2012.06.28"


def open_input_file(input_filename):
    if input_filename.endswith(".gz"):
        import gzip
        handle = gzip.open(input_filename)
    elif input_filename.endswith(".bz2"):
        import bz2
        handle = bz2.BZ2File(input_filename)
    else:
        handle = open(input_filename)
    return handle

def analyze_header(line):
    words = line.strip().split("\t")
    if words[0]!="id":
        raise ValueError("Expected 'id' as first column")
    if words[1]!="chrom":
        raise ValueError("Expected 'chrom' as second column")
    if words[2] not in ("start", "start.0base"):
        raise ValueError("Expected 'start' or 'start.0base' as third column")
    if words[3]!="end":
        raise ValueError("Expected 'end' as fourth column")
    if words[4]!="strand":
        raise ValueError("Expected 'strand' as fifth column")
    if words[5]!="pos":
        raise ValueError("Expected 'pos' as sixth column")
    tpm_index = []
    samples = {'raw': [], 'norm': []}
    words = words[6:]
    n = len(words)
    for i in range(n):
        word = words[i]
        if word.startswith("raw."):
            sample = word[4:]
            samples['raw'].append(sample)
        elif word.startswith("norm."):
            sample = word[5:]
            samples['norm'].append(sample)
            tpm_index.append(i)
        else:
            raise ValueError("Unknown column %s" % word)
    return samples, tpm_index

def read_meta_data(handle):
    assembly = None
    for line in handle:
        if not line.startswith("#"):
            break
        if line.startswith("##ParameterValue[genome_assembly]"):
            key, value = line.strip().split("=", 1)
            assembly = value.strip()
    # current line is the header
    samples, tpm_index = analyze_header(line)
    return assembly, samples, tpm_index

def write_meta_data(output, assembly, basename, samples, window_size):
    # General meta-data
    output.write("##FileFormat = OSCtable\n")
    output.write("##ProtocolREF = OP-CAGE-Tag-Clustering-v1.1\n")
    output.write("##ParameterValue[script_version] = %s\n" % __version__)
    if assembly:
        output.write("##ParameterValue[genome_assembly] = %s\n" % assembly)
    output.write("##ParameterValue[level3_window_size] = %d\n" % window_size)
    output.write("##Date = %s\n" % time.strftime("%Y-%m-%d"))
    output.write('##ContactName = LSA support staff\n')
    output.write('##ContactEmail = lsa-support@gsc.riken.jp\n')
    # Input file name
    output.write("##InputFile = %s\n" % basename)
    # Column variables
    output.write('##ColumnVariable[id] = identifier of this level-3 cluster\n')
    output.write('##ColumnVariable[chrom] = name of the chromosome, identical to what is written in genome assembly\n')
    output.write('##ColumnVariable[start.0base] = start genomic coordinate of a cluster, using zero-based coordinates, always same or smaller than "end"\n')
    output.write('##ColumnVariable[end] = end genomic coodinate of this cluster\n')
    output.write('##ColumnVariable[strand] = genomic strand of this cluster\n')
    output.write('##ColumnVariable[pos] = genomic coordinate of the most highly expressed position of this cluster\n')
    for sample in samples['raw']:
        output.write('##ColumnVariable[raw.%s] = rescued tag count of sample %s\n' % (sample, sample))
    for sample in samples['norm']:
        output.write('##ColumnVariable[norm.%s] = normalized expression by total tag count for sample %s\n' % (sample, sample))
    # Write header line
    line = "id\tchrom\tstart.0base\tend\tstrand\tpos"
    for word in samples['raw']:
        line += "\traw." + word
    for word in samples['norm']:
        line += "\tnorm." + word
    line += "\n"
    output.write(line)

def create_clusters(input_file, tpm_index, window):
    cluster_chromosome = None
    for line in input_file:
        words = line.strip().split('\t')
        name = words[0]
        assert name[:3]=="L2_"
        chromosome = words[1]
        start = int(words[2])
        end = int(words[3])
        assert end > start
        strand = words[4]
        position = int(words[5])
        values = array(words[6:], float)
        sumtpm = sum(take(values, tpm_index))
        if chromosome!=cluster_chromosome or strand!=cluster_strand or start >= cluster_end + window:
            if cluster_chromosome:
                yield (cluster_chromosome, cluster_strand, cluster_tss, cluster_start, cluster_end, cluster_total)
            cluster_chromosome = chromosome
            cluster_strand = strand
            cluster_start = start
            cluster_end = end
            cluster_tss = position
            cluster_total = values
            cluster_maximum = sumtpm
        else:
            cluster_end = end
            cluster_total += values
            if sumtpm > cluster_maximum:
                cluster_maximum = sumtpm
                cluster_tss = position
    yield (cluster_chromosome, cluster_strand, cluster_tss, cluster_start, cluster_end, cluster_total)

def make_line(name, chromosome, strand, tss, start, end, values):
    line = "L3_%s_%s_%s_%d\t%s\t%d\t%d\t%s\t%d" % (name, chromosome, strand, tss, chromosome, start, end, strand, tss)
    for value in values:
        line += "\t" + str(value)
    line += "\n"
    return line

def run(input_filename, output_filename, window, name):
    try:
        input_file = open_input_file(input_filename)
    except IOError:
        print "Failed to open the input file %s" % input_filename
        sys.exit(2)
    if output_filename:
        output = open(output_filename, 'w')
    else:
        output = sys.stdout
    basename = os.path.basename(input_filename)
    assembly, samples, tpm_index = read_meta_data(input_file)
    if name==None:
        if assembly:
            name==assembly
        else:
	    name==""
    write_meta_data(output, assembly, basename, samples, window)
    clusters = create_clusters(input_file, tpm_index, window)
    for cluster in clusters:
        (chromosome, strand, tss, start, end, values) = cluster
        if chromosome==None or strand==None:
            continue
        line = make_line(name, chromosome, strand, tss, start, end, values)
        output.write(line)
    input_file.close()
    if output_filename:
        output.close()


def version():
    print """\
Level-3 promoter clustering version %s.
Copyright (C) 2010-2012 Michiel JL de Hoon (mdehoon@gsc.riken.jp),
RIKEN Omics Science Center, Yokohama, Japan.
    """ % __version__

def usage():
    version()
    print """\
Usage:

python level3.py -i <level2.osc[.gz|.bz2]> [-o <level3.osc>[ [-w <window>]
                 [-n <name>]

     Options:
        -i <level2.osc[.gz|.bz2]>
                           name of the input file containing the level-2
                           promoters in OSC Table format. Optionally, the
                           input file can be gzipped or bzipped.
        -o <level3.osc>    name of the output file with the level-3
                           promoters in OSC Table format.
                           (default: write to stdout)
        -w <window>        the minimum distance in base pairs between
                           adjacent clusters on the genome.
                           (default: 400)
        -n <name>          the identifier to use in promoter names, which
                           are formatted as L3_name_chromosome_strand_position
                           (default: the assembly name as found in the input
                           files)

This script performs level-2 to level-3 promoter clustering.
"""

def main(argv):
    import getopt
    try:
        opts, args = getopt.getopt(argv, "vhi:o:w:n:", ["version", "help", "name"])
    except getopt.GetoptError, error:
        print str(error)
        usage()
        sys.exit(2)
    input_filename = None
    output_filename = None
    window = 400
    name = None
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            usage()
            sys.exit()
        elif opt in ("-v", "--version"):
            version()
            sys.exit()
        elif opt=="-i":
            input_filename = arg
        elif opt=="-o":
            output_filename = arg
        elif opt=="-t":
            threshold = float(arg)
        elif opt=="-w":
            window = int(arg)
        elif opt in ("-n", "--name"):
            name = arg
    if not input_filename:
        usage()
        sys.exit(2)
    run(input_filename, output_filename, window, name)


if __name__ == "__main__":
    main(sys.argv[1:])
